/*

Licensed under the GNU AFFERO GENERAL PUBLIC LICENSE Version 3 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    https://www.gnu.org/licenses/agpl-3.0.html

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package application

import (
	"context"
	"sort"
	"time"

	"github.com/presslabs/controller-util/syncer"
	batchv1 "k8s.io/api/batch/v1"
	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
	"k8s.io/apimachinery/pkg/runtime"
	"k8s.io/apimachinery/pkg/util/wait"
	kstatus "sigs.k8s.io/cli-utils/pkg/kstatus/status"
	"sigs.k8s.io/controller-runtime/pkg/client"

	settings "k8s.libre.sh/application/settings"
	"k8s.libre.sh/application/settings/parameters"
	"k8s.libre.sh/interfaces"
	"k8s.libre.sh/meta"
)

type Instance interface {
	meta.Instance
	// GetComponents() map[int]Component
	GetComponents() map[string]Component
	GetOwner() interfaces.Object
	GetSettings() map[string]settings.Settings
	SetDefaults()
}

func GetTemplateValues(i Instance) map[string]interface{} {
	values := make(map[string]interface{})
	cptsVals := make(map[string]interface{})
	settsVals := make(map[string]interface{})

	for _, cpt := range i.GetComponents() {
		cptsVals[cpt.GetComponent()] = cpt.TemplateValues()
	}

	for _, sett := range i.GetSettings() {
		// TODO create templateValues for Settings interface
		settsVals[sett.GetMeta().GetComponent()] = parameters.KeyPairValues(sett.GetConfig().GetParameters())
	}

	values["components"] = cptsVals
	values["settings"] = settsVals

	return values

}

func CreateAndInitSettings(i Instance, r interfaces.Reconcile) (setts map[string]settings.Settings, err error) {

	setts = i.GetSettings()

	// First we init external values and  random values
	for _, sett := range setts {
		err = sett.Init(r.GetClient())

		if err != nil {
			return setts, err
		}

	}

	// Transform instance to template values used to init settings
	values := GetTemplateValues(i)

	// Init the values from templates
	for _, sett := range setts {
		sett.GetConfig().InitTemplateValues(values)
	}

	return setts, err
}

func NewSyncers(app Instance, r interfaces.Reconcile, owner interfaces.Object) (syncers []syncer.Interface, err error) {
	return syncers, nil
}

// TODO Transform to createandinit and return application object so we can store the values
func Init(i Instance, r interfaces.Reconcile) error {
	return nil
}

func Sync(r interfaces.Reconcile, syncers map[string]map[int]syncer.Interface, order map[int]string) error {
	keys := []int{}

	for k := range order {
		keys = append(keys, k)
	}

	sort.Ints(keys)

	for _, k := range keys {
		ss := syncers[order[k]]

		objKeys := []int{}

		for kk := range ss {
			objKeys = append(objKeys, kk)
		}

		for _, kk := range objKeys {
			if err := syncer.Sync(context.TODO(), ss[kk], r.GetRecorder()); err != nil {
				return err
			}

			err := WaitUntilCompletion(r, ss[kk].(*syncer.ObjectSyncer).Obj)

			if err != nil {
				//		r.Log.Error(err, "Migration job failed")
				//	r.Recorder.Event(&app, "Normal", fmt.Sprintf("%s", op), fmt.Sprintf("Job migration %s/%s", app.Namespace, app.Name))
				return err
			} else {
				//		r.Log.Info("Migration job is completed")
				//	r.Recorder.Event(&app, "Normal", fmt.Sprintf("%s", op), fmt.Sprintf("Job migration completed %s/%s", app.Namespace, app.Name))
			}
		}
	}

	return nil
}

func WaitUntilCompletion(r interfaces.Reconcile, obj runtime.Object) error {

	return wait.Poll(RetryInterval, ReadinessTimeout, func() (bool, error) {
		objectKey, _ := client.ObjectKeyFromObject(obj)

		if err := r.GetClient().Get(context.Background(), objectKey, obj); err != nil {
			return false, err
		}

		unstruct := &unstructured.Unstructured{}
		unstructuredObj, err := runtime.DefaultUnstructuredConverter.ToUnstructured(obj)

		if err != nil {
			return false, err
		}

		unstruct.Object = unstructuredObj

		s, err := kstatus.Compute(unstruct)

		// Check if job then check if completed
		obj, ok := obj.(*batchv1.Job)
		if ok {
			if obj.Status.Succeeded > 0 { // || job.Status.Failed > types.Int32(job.Spec.BackoffLimit)  {
				return true, nil
			}
			return false, nil

		}

		if err != nil {
			return false, err
		}
		if s.Status == kstatus.FailedStatus {
			return false, err
		}
		if s.Status == kstatus.UnknownStatus {
			return false, err
		}
		if s.Status == kstatus.InProgressStatus {
			return false, nil
		}

		return true, nil
	})
}

const (
	RetryInterval    = 50 * time.Millisecond
	RetryTimeout     = 2 * time.Second
	ReadinessTimeout = 10 * time.Minute
	GCTimeout        = 5 * time.Minute
)
